@@ -5,7 +5,7 @@ package OpenAPI::Modern;
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.018';
+our $VERSION = '0.019';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -76,20 +76,19 @@ around BUILDARGS => sub ($orig, $class, @args) {
   return $args;
 };
 
-# at the moment, we rely on these values being provided in $options:
-# - path_template OR operationId
-# - path_captures
 sub validate_request ($self, $request, $options = {}) {
   my $state = {
     data_path => '/request',
     initial_schema_uri => $self->openapi_uri,   # the canonical URI as of the start or last $id, or the last traversed $ref
     traversed_schema_path => '',    # the accumulated traversal path as of the start, or last $id, or up to the last traversed $ref
     schema_path => '',              # the rest of the path, since the last $id or the last traversed $ref
-    errors => [],
+    errors => $options->{errors} //= [],
   };
 
   try {
-    my ($path_template, $path_captures) = $self->_find_path($state, $request, $options);
+    die pop $options->{errors}->@* if not $self->find_path($request, $options);
+
+    my ($path_template, $path_captures) = $options->@{qw(path_template path_captures)};
     my $path_item = $self->openapi_document->schema->{paths}{$path_template};
     my $method = lc $request->method;
     my $operation = $path_item->{$method};
@@ -118,7 +117,9 @@ sub validate_request ($self, $request, $options = {}) {
         next if exists $request_parameters_processed->{$param_obj->{in}}{$fc_name};
         $request_parameters_processed->{$param_obj->{in}}{$fc_name} = $section;
 
-        $state->{data_path} = jsonp($state->{data_path}, $param_obj->{in}, $param_obj->{name});
+        $state->{data_path} = jsonp($state->{data_path},
+          ((grep $param_obj->{in} eq $_, qw(path query)) ? 'uri' : ()), $param_obj->{in},
+          $param_obj->{name});
         my $valid =
             $param_obj->{in} eq 'path' ? $self->_validate_path_parameter($state, $param_obj, $path_captures)
           : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter($state, $param_obj, $request->uri)
@@ -131,7 +132,8 @@ sub validate_request ($self, $request, $options = {}) {
     # 3.2 "Each template expression in the path MUST correspond to a path parameter that is included in
     # the Path Item itself and/or in each of the Path Item’s Operations."
     foreach my $path_name (sort keys $path_captures->%*) {
-      abort($state, 'missing path parameter specification for "%s"', $path_name)
+      abort({ %$state, data_path => jsonp($state->{data_path}, qw(uri path), $path_name) },
+          'missing path parameter specification for "%s"', $path_name)
         if not exists $request_parameters_processed->{path}{$path_name};
     }
 
@@ -163,22 +165,23 @@ sub validate_request ($self, $request, $options = {}) {
     }
   }
 
+  $options->{errors} = $state->{errors};
   return $self->_result($state);
 }
 
-# at the moment, we rely on these values being provided in $options:
-# - path_template OR operationId
 sub validate_response ($self, $response, $options = {}) {
   my $state = {
     data_path => '/response',
     initial_schema_uri => $self->openapi_uri,   # the canonical URI as of the start or last $id, or the last traversed $ref
     traversed_schema_path => '',    # the accumulated traversal path as of the start, or last $id, or up to the last traversed $ref
     schema_path => '',              # the rest of the path, since the last $id or the last traversed $ref
-    errors => [],
+    errors => $options->{errors} //= [],
   };
 
   try {
-    my ($path_template, $path_captures) = $self->_find_path($state, $response->request, $options);
+    die pop $options->{errors}->@* if not $self->find_path($response->request, $options);
+
+    my ($path_template, $path_captures) = $options->@{qw(path_template path_captures)};
     my $method = lc $response->request->method;
     my $operation = $self->openapi_document->schema->{paths}{$path_template}{$method};
 
@@ -229,27 +232,30 @@ sub validate_response ($self, $response, $options = {}) {
     }
   }
 
+  $options->{errors} = $state->{errors};
   return $self->_result($state);
 }
 
-######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
-
-# at the moment, we rely on these values being provided in $options:
-# - path_template OR operation_id
-# return: path_template.  caller can get method out of $request.
-# in the future, we will parse the request URI and can extract path_captures as well.
-sub _find_path ($self, $state, $request, $options) {
+sub find_path ($self, $request, $options) {
   my $path_template;
 
+  my $state = {
+    data_path => '/request/uri/path',
+    initial_schema_uri => $self->openapi_uri,   # the canonical URI as of the start or last $id, or the last traversed $ref
+    traversed_schema_path => '',    # the accumulated traversal path as of the start, or last $id, or up to the last traversed $ref
+    schema_path => '',              # the rest of the path, since the last $id or the last traversed $ref
+    errors => $options->{errors} //= [],
+  };
+
   # path_template from options, method from request
   if (exists $options->{path_template}) {
     $path_template = $options->{path_template};
 
     my $path_item = $self->openapi_document->schema->{paths}{$path_template};
-    abort({ %$state, keyword => 'paths' }, 'missing path-item "%s"', $path_template) if not $path_item;
+    return E({ %$state, keyword => 'paths' }, 'missing path-item "%s"', $path_template) if not $path_item;
 
     my $method = lc $request->method;
-    abort({ %$state, schema_path => jsonp('/paths', $path_template), keyword => $method },
+    return E({ %$state, data_path => '/request/method', schema_path => jsonp('/paths', $path_template), keyword => $method },
         'missing entry for HTTP method "%s"', $method)
       if not $path_item->{$method};
   }
@@ -257,17 +263,18 @@ sub _find_path ($self, $state, $request, $options) {
   # path_template and method from operationId from options
   if (exists $options->{operation_id}) {
     my $operation_path = $self->openapi_document->get_operationId($options->{operation_id});
-    abort({ %$state, keyword => 'paths' }, 'unknown operation_id "%s"', $options->{operation_id})
+    return E({ %$state, keyword => 'paths' }, 'unknown operation_id "%s"', $options->{operation_id})
       if not $operation_path;
-    abort({ %$state, schema_path => $operation_path, keyword => 'operationId' },
+    return E({ %$state, schema_path => $operation_path, keyword => 'operationId' },
       'operation id does not have an associated path') if $operation_path !~ m{^/paths/};
     (undef, undef, $path_template, my $method) = unjsonp($operation_path);
 
-    abort({ %$state, schema_path => jsonp('/paths', $path_template) },
+    return E({ %$state, schema_path => jsonp('/paths', $path_template) },
         'operation does not match provided path_template')
       if exists $options->{path_template} and $options->{path_template} ne $path_template;
 
-    abort({ %$state, schema_path => $operation_path }, 'wrong HTTP method %s', $request->method)
+    return E({ %$state, data_path => '/request/method', schema_path => $operation_path },
+        'wrong HTTP method %s', $request->method)
       if lc $request->method ne $method;
   }
 
@@ -286,17 +293,17 @@ sub _find_path ($self, $state, $request, $options) {
       my @capture_values = map
         Encode::decode('UTF-8', URI::Escape::uri_unescape(substr($uri_path, $-[$_], $+[$_]-$-[$_]))), 1 .. $#-;
       my @capture_names = ($path_template =~ m!\{([^/?#}]+)\}!g);
-      my %path_captures; @path_captures{@capture_names} = map URI::Escape::uri_unescape($_), @capture_values;
-      return ($path_template, \%path_captures);
+      my %path_captures; @path_captures{@capture_names} = @capture_values;
+      $options->@{qw(path_template path_captures)} = ($path_template, \%path_captures);
+      return 1;
     }
 
-    abort({ %$state, keyword => 'paths', data_path => jsonp($state->{data_path}, 'path') },
-      'no match found for URI path "%s"', $uri_path);
+    return E({ %$state, keyword => 'paths' }, 'no match found for URI path "%s"', $uri_path);
   }
 
   # note: we aren't doing anything special with escaped slashes. this bit of the spec is hazy.
   my @capture_names = ($path_template =~ m!\{([^/}]+)\}!g);
-  abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+  return E({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
       'provided path_captures names do not match path template "%s"', $path_template)
     if exists $options->{path_captures}
       and not is_equal([ sort keys $options->{path_captures}->%*], [ sort @capture_names ]);
@@ -304,22 +311,25 @@ sub _find_path ($self, $state, $request, $options) {
   # 3.2: "The value for these path parameters MUST NOT contain any unescaped “generic syntax”
   # characters described by [RFC3986]: forward slashes (/), question marks (?), or hashes (#)."
   my $path_pattern = $path_template =~ s!\{[^/}]+\}!([^/?#]*)!gr;
-  abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+  return E({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
       'provided %s does not match request URI', exists $options->{path_template} ? 'path_template' : 'operation_id')
     if $uri_path !~ m/^$path_pattern$/;
 
   # perldoc perlvar, @-: $n coincides with "substr $_, $-[n], $+[n] - $-[n]" if "$-[n]" is defined
   my @capture_values = map
     Encode::decode('UTF-8', URI::Escape::uri_unescape(substr($uri_path, $-[$_], $+[$_]-$-[$_]))), 1 .. $#-;
-  abort({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
+  return E({ %$state, keyword => 'paths', _schema_path_suffix => $path_template },
       'provided path_captures values do not match request URI')
     if exists $options->{path_captures}
       and not is_equal([ map $_.'', $options->{path_captures}->@{@capture_names} ], \@capture_values);
 
   my %path_captures; @path_captures{@capture_names} = @capture_values;
-  return ($path_template, \%path_captures);
+  $options->@{qw(path_template path_captures)} = ($path_template, \%path_captures);
+  return 1;
 }
 
+######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
+
 sub _validate_path_parameter ($self, $state, $param_obj, $path_captures) {
   # 'required' is always true for path parameters
   return E({ %$state, keyword => 'required' }, 'missing path parameter: %s', $param_obj->{name})
@@ -543,13 +553,13 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI docume
 
 =head1 VERSION
 
-version 0.018
+version 0.019
 
 =head1 SYNOPSIS
 
   my $openapi = OpenAPI::Modern->new(
     openapi_uri => 'openapi.yaml',
-    openapi_schema => YAML::PP->new(boolean => 'JSON::PP')->load_string(<<'YAML'),
+    openapi_schema => YAML::PP->new(boolean => 'JSON::PP')->load_string(<<'YAML'));
   openapi: 3.1.0
   info:
     title: Test API
@@ -597,17 +607,13 @@ version 0.018
                     status:
                       const: ok
   YAML
-  );
 
   say 'request:';
   use HTTP::Request::Common;
   my $request = POST 'http://example.com/foo/bar',
     [ 'My-Request-Header' => '123', 'Content-Type' => 'application/json' ],
     '{"hello": 123}';
-  say $openapi->validate_request($request, {
-    path_template => '/foo/{foo_id}',
-    path_captures => { foo_id => 'bar' },
-  });
+  say $openapi->validate_request($request);
 
   say 'response:';
   my $response = HTTP::Response->new(
@@ -615,7 +621,7 @@ version 0.018
     [ 'My-Response-Header' => '123' ],
     '{"status": "ok"}',
   );
-  say $openapi->validate_response($response, { operation_id => 'my_foo_request' });
+  say $openapi->validate_response($response);
 
 prints:
 
@@ -694,6 +700,26 @@ The L<JSON::Schema::Modern> object to use for all URI resolution and JSON Schema
 Validates an L<HTTP::Request> object against the corresponding OpenAPI v3.1 document, returning a
 L<JSON::Schema::Modern::Result> object.
 
+The second argument is a hashref that contains extra information about the request, corresponding to
+the values expected by L</find_path> below. It is populated with some information about the request:
+pass it to a later L</validate_response> to improve performance.
+
+=head2 validate_response
+
+  $result = $openapi->validate_response(
+    $response,
+    {
+      path_template => '/foo/{arg1}/bar/{arg2}',
+    },
+  );
+
+The second argument is a hashref that contains extra information about the request, as in L</find_path>.
+
+=head2 find_path
+
+  $result = $self->find_path($request, $options);
+
+Uses information in the request to determine the relevant parts of the OpenAPI specification
 The second argument is a hashref that contains extra information about the request. Possible values include:
 
 =over 4
@@ -716,21 +742,15 @@ All of these values are optional, and will be derived from the request URI as ne
 efficiently than if they were provided). All passed-in values MUST be consistent with each other and
 the request URI.
 
+When successful, the options hash will be populated with keys C<path_template> and C<path_captures>
+and the return value is true.
+When not successful, the options hash will be populated with key C<errors>, an arrayref containing
+a L<JSON::Schema::Modern::Error> object, and the return value is false.
+
 Note that the L<C</servers>|https://spec.openapis.org/oas/v3.1.0#server-object> section of the
 OpenAPI document is not used for path matching at this time, for either scheme and host matching nor
 path prefixes.
 
-=head2 validate_response
-
-  $result = $openapi->validate_response(
-    $response,
-    {
-      path_template => '/foo/{arg1}/bar/{arg2}',
-    },
-  );
-
-The second argument is a hashref that contains extra information about the request, as in L</validate_request>.
-
 =head2 canonical_uri
 
 An accessor that delegates to L<JSON::Schema::Modern::Document/canonical_uri>.
